/* 0927, Mon 30 Aug 10 (NZST)
   1550, Tue 27 Jul 10 (CEST)
   1253, Wed 19 May 10 (NZST)
   1955, Thu 22 Apr 10 (PDT)
   1445, Wed  5 Dec 07 (NZDT)

   owt.h:  Globals for OneWayTraffic Monitor

   Copyright (C) 2008-2010 by Nevil Brownlee, U Auckland | CAIDA | WAND

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>
*/

include <sys/types.h>
#include <sys/stat.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <syslog.h>
#include <unistd.h>

#include <pthread.h>
#include "semaphore.h"

#include "libtrace.h"

#include "../owt_conf.h"

/* odin:      SIXTYFOURBIT 1, DFMXFLOWWATCHERS 1, STREAMTIMEOUTS 0
   ua-trace:  SIXTYFOURBIT 1, DFMXFLOWWATCHERS 1, STREAMTIMEOUTS 1
   ua-live:   SIXTYFOURBIT 0, DFMXFLOWWATCHERS 4, STREAMTIMEOUTS 1

   Fri, 21 May 10 (NZST):  These are set by ./configure in owt_conf.h */


// #define SIXTYFOURBIT       1  /* Odin is a 64-bit machine */
// #define DFMXFLOWWATCHERS   1  /* Number of flow watcher threads */
//                              /*   (4 has more overhead on odin) */
// #define STREAMTIMEOUTS     0  /* Time out inactive (one-way) streams */

#define MKSTATSCOUNTS      0  /* To build/collect interval stats */
#define MKDESTCOUNTS       1  /* To build/collect dst_addr/24block stats */
#define NDESTBINS        256  /* 256 /24 blocks in telescope space */

#define CHECKPKTQUEUE      0  /* Debugging defines */
#define STOP_AFTER         0  /* stop_meter() after this many #Time calls */
                              /*   0 to only stop at EOF */
#define SC_DEBUG  0  /* Stream chain checks:
         0 = none
         1 = start of cto() [check_stream_timeouts()]
         2 = start cto() + print cto() stats
         3 = start & end of cto() + end of each_pkt() */
#define STARTUP_DEBUG  0  /* 1 to print 'got to here' lines */


#define PRODUCTION         1  /* 0 to display statistics < every minute */
#if PRODUCTION
#define STATS_RECORDING_INTERVAL  60  /* 1m */
#else
#define STATS_RECORDING_INTERVAL  20  /* 20s */
#endif
#define STATS_INTERVALS_PER_HOUR  3600/STATS_RECORDING_INTERVAL
/* #Time recording interval is set by -c in owt.init method */

#define METER_VERSION  "1.1"  /* 1.1 for UA with two-way timeouts */

#include "owt_queue.h"
                               /* Default sizes */
#define DFMXOPQENTRIES  25000  /* Size of output queue */
#define DFMXSTREAMS    250000  /* Nbr of streams */
#define DFMXEVENTS         10  /* Nbr of events */
#define DFMXSTREAMTXTS     20  /* Nbr of 'stream txt' messages */
#define DFMXPWQUEUESZ    1000  /* Max length of packet watcher input queue */
#define MXSTRMTXTLEN      501  /* Max length of a 'stream txt' message */
#define DFMXPKTDATAS  1000000  /* Nbr of pkt_data objects */
#define DFPKTSTORECORD     10  /* Pkts to record per stream */
#define DFTRACESTOPROCESS   1  /* Number of trace files to process */

#define MXINTERFACES        2  /* Max interfaces we can watch */

#define MIN_HT_SIZE   2000000  /* Don't want long hash chains! */

#include "owt_mem.h"  /* Stream_engine memory management defines */

#define IP4_ADDR_LEN      4  /* IP address sizes */
#define IP6_ADDR_LEN     16

#define PT_ICMP           1  /* IP protocol types */
#define PT_TCP            6
#define PT_UDP           17
#define PT_SCTP         132

#define ICMP_DEST_UNREACHABLE   3  /* ICMP Types */
#define ICMP_TTL_EXCEEDED      11
#define ICMP6_DEST_UNREACHABLE  1  /* ICMP6 Types */
#define ICMP6_TTL_EXCEEDED      3

#define SKEY_OK           0  /* s_key return codes */
#define SKEY_NOSPACE      1
#define SKEY_NOTIP        2
#define SKEY_NOPEERADDR   3
#define SKEY_FRAGPKT      4
#define SKEY_NOTRANSADDR  5
#define SKEY_BADIPHDR     6
#define SKEY_V6SMALLSNAP  7
#define SKEY_V6OPAQUE     8
#define SKEY_NOLOCALADDR  9

extern char *logname;
void log_msg(int priority, int die, char *fmt, ...);


struct owt_ht {
   int table_size;  /* Max entries, used as hashmod */
   pthread_mutex_t ht_lock;  /* Lock for hash table */
   void **sfht;  /* Actual hash table */
   };

#define SS_FREE          0  /* Stream states */
#define SS_NEW           1
#define SS_TWO_WAY       2  /* Don't need to save pkts */
#define SS_ENOUGH        3  /* We've saved enough pkts */
#define SS_DELETE        4
#define SS_TIMEOUT       5
#define SS_n_states      6

#pragma pack(push,1)

struct ip_address {
   union a {
      uint32_t v4;
      uint8_t a[IP6_ADDR_LEN];
      struct o {
         uint64_t oct0, oct1;
         } o;
      } a;
   uint8_t ver, len;
   };

struct pkt_data {
   struct pkt_data *next;
   double arrival_time;
   uint8_t proto, tcp_flags, cfc, last_op;
   struct ip_address dst_addr;
   uint16_t src_port, dst_port,
      ip_len, res16;
   };
                             /* last_op values */
#define PD_FREE          0   /* free_stream */
                             /* 1-5 are the SS states */
#define PD_READ    SS_n_states   /* by get_v*key */
#define PD_n_states  PD_READ+1

owt_queue_declare_globals(pkt_data);
owt_mem_declare_globals(pkt_data);

struct stream {  /* Hash table entry, key is (src/dest addr) */
   struct stream *next_hc;    /* For utht(), must be first */
   uint8_t version, state,    /* For utht(), should be second */   
      first_fwd,  /* Remembers fwd value for first packet */
      timeout;    /* In seconds for this stream */

   struct stream *next, *prev;  /* For stream queue */
   struct ip_address src_addr;

   uint32_t tp, fp,           /* To/From packet counts */
     tb, fb,                  /* To/From byte counts */
     tcp_pkts, udp_pkts, other_pkts;  /* For per-source-type stats */
   uint64_t totp;             /* Total packet count */
   double first_us, last_us;  /* Times for first and last packets */
   uint32_t last_report_sec;  /* ts when last stream report was queued */

   /* Data for individual packet (or event) .. */
   uint8_t event_type,  /* pkt event (_EVENT values, 0 to just count pkt) */
      fwd;              /* fwd value for this packet */
   uint16_t wl;         /* Packet's wire length */
   uint32_t ts;         /* Seconds field of packet's timestamp */
   uint32_t hash;       /* 32-bit hash for sp */

   struct pkt_data *pktp;  /* For incoming packets */
   owt_queue_declare_space(pktq, pkt_data);  /* Packet data queue */

   uint32_t probe_port;
   double enough_lifetime, retry_time;
   uint16_t pkt_len_mode;
   uint8_t src_type, probe_port_valid, analysed, n_dest_ports,
     last_active_interval,
     ext_blk,  /* 0 = stream_space, >0 = extra_streams[] */
     fw_nbr, pkts_last_classify,
     active_since_last_sum, msg_queued;
   };

owt_queue2_declare_globals(stream);
owt_mem_declare_globals(stream);

#define MX_EXTRA_STREAMS         7
extern struct stream *extra_streams[MX_EXTRA_STREAMS+1];
extern int extra_stream_incr, n_extra_streams;

#pragma(pop)

#if MKSTATSCOUNTS
extern uint32_t  /* Per-interval stats arrays for each source type .. */
//x   **pkts_seen_per_stats_interval, *pkts_seen_this_interval,
   **active_sources_per_stats_interval, *active_sources_this_interval,
   **unique_hosts_per_stats_interval, *unique_hosts_this_interval;
extern uint32_t stats_interval_nbr;
#endif
#if MKDESTCOUNTS
extern uint32_t **dest_addrs_per_source_type;
#define src_addr_bin(a)  ((ntohl(a) >> 24) & 0x000000FF)  /* 1st byte */
extern uint32_t (*dest_addr_bin)(uint32_t a);
#endif

#define MX_LOCAL_ADDRS  5
extern struct ip_address la[MX_LOCAL_ADDRS];
extern int n_local_addrs;

#define TCP_STREAM_TIMEOUT     2  /* 2s (FIN_WAIT) */
#define OTHER_STREAM_TIMEOUT   5  /* 5s */
#define STO_MULTIPLIER        10

#define MX_ONE_WAY_TIMEOUTS_PER_CHECK  20000  /* 6 Aug 10 */
   /* Limit demand on opq_empty!  23 Jun 10 */

#define TWO_WAY_TIMEOUT       OTHER_STREAM_TIMEOUT
#define ONE_WAY_TIMEOUT       10  /* 10s */

#define STO_TWO_WAY           2  /* Timeout kinds for streams .. */
#define STO_ONE_WAY           1

extern int n_source_types, pkts_to_record, min_pkts_to_classify,
   min_pkts_for_active_source, traces_to_process;
extern char **other_uris;
extern int next_uri_nbr;


extern int time_rec_interval, intervals_this_sum_file;

extern int n_lost_stream_v4s, n_lost_stream_v6s;
extern uint64_t last_drops;

extern uint32_t mx_active_streams, mx_active_pkt_datas,
   mx_active_opq_entries,
   av_cl_count, n_hc_searches,  src_av_cl_count, src_n_hc_searches;
extern uint32_t n_not_ip,n_frags,n_too_small,n_no_local_address,
   n_v6_opaque;

extern uint64_t one_way_bytes, two_way_bytes,  /* Overall stats */
   one_way_pkts, two_way_pkts,
   one_way_streams, two_way_streams,
   v6_one_way_bytes, v6_two_way_bytes,
   v6_one_way_pkts, v6_two_way_pkts,
   v6_one_way_streams, v6_two_way_streams,
   not_one_way_streams;

extern uint32_t start_ts;  /* Seconds part of first-packet timestamp */

struct if_info {
   uint32_t n_pkts,  /* Packets this second */
      n_seconds,     /* Seconds since start of interval */
      mx_pkt_rate, av_pkt_count;  /* Max/s, total pkts this interval */
   uint32_t last_ts;
   char *if_name;
   int thread_running;
   pthread_t pw_thread;  /* Thread for input packet watcher */
   libtrace_t *trace;
   libtrace_packet_t *packet;
   int if_nbr;
   };

extern int snaplen;
extern uint32_t g_last_ts;  /* Max last_ts for all interfaces */

extern struct if_info if_info[MXINTERFACES];
extern int n_interfaces,   /* Nbr of interfaces specified */
   n_interfaces_running;

#define OQ_NULL           0  /* OQ (Output Queue) entry types */
#define OQ_TXTMSG         1
#define OQ_STARTED        2  /* Not used 21 Aug 10 */
#define OQ_PROGRESS       3  /* Not used 21 Aug 10 */
#define OQ_ENOUGH         4
#define OQ_TIMED_OUT      5
#define OQ_END_OF_FILE    6

struct opq_entry {
   uint8_t type;  /* OQ_ type */
   int s_id;      /* Index in stream or src_host array */
   char *msg;     /* For 'stream txt' message */
   time_t ts;     /* ts seconds for txt msg */
   struct opq_entry *next; 
   };

owt_queue_declare_globals(opq_entry);
owt_mem_declare_globals(opq_entry);

extern owt_queue_declare_space(opq, opq_entry);  /* owt output queue */

#define NO_EVENT       0  /* Event types */
#define S_TIME_EVENT   1  /* Check stream timeouts */
#define STATS_EVENT    2  /* Write interface stats */
#define TIME_EVENT     3  /* Write #Time record */
#define END_OF_FILE    4  /* End of trace file */

struct event {
   uint32_t t;
   uint16_t type, interval;  /* Interval in seconds */
   struct event *next;
   };
extern struct event event_q;  /* Queue of active events, dummy first member */

owt_queue_declare_globals(event);
owt_mem_declare_globals(event);

struct stream_txt {
   char msg[MXSTRMTXTLEN+1];
   struct stream_txt *next;
   };

owt_queue_declare_globals(stream_txt);
owt_mem_declare_globals(stream_txt);

/* mutexes for data structures shared between threads .. */
extern pthread_mutex_t 
   log_file_mutex,     /* log_file */
   event_q_mutex,      /* Event queue */
   if_info_mutex,      /* if_info table */
   stream_data_mutex,  /* stream data (packet/byte counts) */
   n_fw_mutex,         /* n_fw_running */
   count_mutex;        /* lock for count/count_cv */

extern  pthread_cond_t count_cv;

extern int count;  /* Nbr of check_timeouts() running */
#define CV_CHECKING   (count > 0)
#define CV_IDLE       (count == 0)
#define CV_RECORDING  (count < 0)

struct flow_watcher_control {
   struct flow_watcher_control *next;
   owt_queue_declare_space(ispq, stream);  /* Packet-watcher output queue */
   Semaphore ispq_ready, ispq_empty;  /* Flow watcher input queue */
  //n   struct owt_ht *src_table;  /* Src_host (hash) table */
   struct owt_ht *sh_table;  /* Stream (hash) table */
   owt_queue2_declare_space(sq, stream);  /* (Bi-directional) stream queue */
   struct stream *eof_sp;  /* So packet-watcher can signal EOF */
   pthread_t flow_thread;
   int id;
   };

owt_queue_declare_globals(flow_watcher_control);
owt_mem_declare_globals(flow_watcher_control);

extern int n_fw_running;  /* Nbr of flow watchers running */

extern Semaphore stream_empty;  /* Controls for packet-watcher input queue */

extern Semaphore opq_ready, opq_empty;  /* Output queue (to outer-block loop) */

#include <sys/utsname.h>
extern struct utsname node;  /* Info about the host we're nunning on */

extern char *version;

/* Functions in owt_watcher.c */

void quack(int nnn);
int sv4p_bad(struct stream *sp);

char *owt_file_header(void);

void update_interface_stats(uint32_t not_ts);
void each_time(uint32_t now_ts);
void check_stream_timeouts(struct flow_watcher_control *fw, uint32_t now_ts);
void output_active_streams(struct stream_queue *sq, uint32_t now_ts);

void one_second_process(uint32_t ts);

extern int request_shutdown;  /* True to stop stream engine */
void reset_summary_info(void);
void start_meter(char *interfaces[], int n_ifs,
   int mx_events, int mx_opq_entries, 
   int mx_pkt_datas, int mx_streams,
   int mx_pw_q_sz, int mx_stream_txts, int mx_flow_watchers,
   int mx_source_types, int hours_reqd);
void stop_meter(char *msg);

void each_pkt(struct flow_watcher_control *fw, struct stream *isp);
void cancel_packet_watchers(void);

#include <unistd.h>

#define SNAP_LEN  120

struct ether_hdr {  
   uint8_t dest[6];    /* MAC addresses */
   uint8_t source[6];
   uint8_t type[2];    /* Length or protocol type */
   };

extern int live_source;  /* Zero for trace files */

double at_seconds(struct timeval tv);
uint32_t pkt_drops(void);
int init_trace(struct if_info *iip);
void close_trace(struct if_info *iip);
void *packet_watcher_process(void *arg);
void *flow_watcher_process(void *arg);

#include <signal.h>
void sigint_handler(int x);

extern int errno;
void usage(char *argv0);

#include <getopt.h>
int main(int argc, char *argv[]);

char *key_to_s(struct stream *sp);

/* Functions in owt_monitor.c */

/* Functions in owt_support.c */

#include <pwd.h>
void uid_from_cwd(void);
extern int my_uid, ut_uid, my_gid, ut_gid;

#include <stdarg.h>

struct event *get_event(int t, int interval);
void schedule_event(struct event *ev, uint32_t first_t);
int check_events(uint32_t now);

struct opq_entry *oqe_for_stream(struct stream *sp);
struct stream *stream_from_oqe(struct opq_entry *oqe);

struct stream *get_stream(int kind);
void free_stream(struct stream *sp);
void bump_way_counts(struct pkt_data *pkt);
void free_pkt_data(struct pkt_data *pktp);
void free_pkt_queue(struct pkt_data_queue *pq);
//n void free_src_host(struct src_host *shp);

void init_mutex(pthread_mutex_t *m, char *name);
void _mutexes(void);

#define mutex_lock(caller,m)  tw_pthread_mutex_lock(m)
#define mutex_unlock(caller,m)  tw_pthread_mutex_unlock(m)
#define ntm_semaphore_down(caller,s)  t_semaphore_down(s)
#define ntm_semaphore_up(caller,s)  t_semaphore_up(s)
#define ntm_semaphore_init(name,id,s)  t_semaphore_init(s)

void owt_start(int mx_events, int mx_opq_entries, 
	       //n   int mx_v4_tuples, int mx_v6_tuples,
   int mx_streams, int mx_pkt_datas,
   int mx_stream_txts, int mx_flow_watchers);
void owt_shutdown(void);

#define v6cmp(s, d)  memcmp(s.a, d.a, IP6_ADDR_LEN)
char *v6addr_to_s(char *v6a, struct ip_address *addrp);

struct owt_ht *owt_ht_create(int mxhtentries);
uint32_t owt_hash(struct stream *record);
void *owt_ht_lookup(struct owt_ht *t, uint32_t hash,
   struct stream *record, int *OK_to_add);
   /* returns record, NULL if not found */
void *owt_ht_delete(struct owt_ht *t, void *record);
   /* deletes specified record, returns it or NULL */

int addr_is_local(int ver, struct ip_address *addrp);

void stream_fl_check(void);
