#!/usr/bin/env ruby

# 1354, Wed  6 Jan 10 (NZDT)
# 1620, Tue, 17 Jun 08 (NZST)
#
# distrib.rb: Distribution class, builds distributions (linear or log),
#             finds means and percentiles, prints distributions to a file
#
# Copyright 2008-2010, Nevil Brownlee, U Auckland | CAIDA | WAND

# Distribution transform types  (@type values)
DT_LINEAR   = 1  # Linear bins, @bins doesn't include overflow bin
DT_LOG      = 2  # Logarithmic bins
DT_DYNAMIC  = 4  # Request dynamic distrib (in an SRL program)
DT_DYN_VALS = 5  # Dynamic, @bins = nbr of actual values
DT_DYN_LIN  = 6  # Linear bins, from a dynamic distribution
DT_DYN_LOG  = 7  # Log bins, dito

class Distribution
  def initialize(lower, upper, bins, type, xscale)
    @lower = lower.to_f;  @upper = upper.to_f;  @bins = bins.to_i
    @type = type;  @xscale = xscale.to_f

    case @type
    when DT_LINEAR
      @M = (@bins-1)/(@upper - @lower)
    when DT_LOG
      @M = (@bins-1)/Math.log10(@upper/@lower)
    end
    # print "distrib.new: lower=#{@lower}, upper=#{@upper}, bins=#{@bins}, type=#{@type}, M=#{@M}\n"
    @counts = Array.new(@bins+1, 0)  # 1 extra for overflow bin
    @n_counts = 0;  @total_value = 0.0
  end

  attr_reader :n_counts

  def import_counts(a)
    if a.size != @bins+1
      print "@bins=#{@bins}, a.size=#{a.size}\n"
      $stderr.print "import_counts - wrong nbr of bins <<<<\n"
      exit
    else
      @n_counts = 0
      a.each_index do |x|
        @counts[x] = a[x].to_i
        @n_counts += @counts[x]
      end
    end
  end

  def count(x)
    @n_counts += 1;  @total_value += x

    if x <= @lower
      n = 0
    elsif x > @upper
      n = @bins
    else
      case @type
      when DT_LINEAR
        @T = (x - @lower) * @M
      when DT_LOG
        @T = Math.log10(x/@lower) * @M
      end
      n = 1 + @T.floor
    end
    @counts[n] += 1
  end

  def mean
    @total_value*@xscale/@n_counts
  end

  def percentile(pc)
    case @type
    when DT_LINEAR
      step = (@upper-@lower)/(@bins-1)
    when DT_LOG
      step =(Math.log10(@upper)-Math.log10(@lower))/(@bins-1)
    end

    t = lp = p = j = 0;  lj = -1  # Find percentile's index into the bins
    @counts.each do |c|
      if c > 0
        t += c;  p = t*100.0/@n_counts
        break if p >= pc
        lp = p;  lj = j
      end
      j += 1
    end
    if lj >= 0
      pv = j - (j-lj)*(p-pc)/(p-lp)
    else  # No counts below pc
      pv = j;
    end

    case @type  # Scale bin index to get percentile's value
    when DT_LINEAR
      x = @lower + pv*step
    when DT_LOG
      x = @lower * 10**(pv*step)
    end
    x*@xscale
  end  # percentile

  PRboth = 0
  PRxonly = 1
  PRyonly = 2
  PRcumuly = 3
  PRxreals = 4  # return array of reals

  def write(of, pc, which)  # pc => dump percentages rather than counts
    case @type
    when DT_LINEAR
      step = (@upper-@lower)/(@bins-1)
    when DT_DYN_LOG, DT_LOG
      step =(Math.log10(@upper)-Math.log10(@lower))/(@bins-1)
    end

    totalv = 0.0
    (0..@bins).each do |n|
      #print "#{@counts[n]} "
      case @type
      when DT_LINEAR
        x = @lower + n*step
      when DT_LOG
        x = @lower * 10**(n*step)
      end
      if pc
        cv = @counts[n]*100.0/@n_counts
      else
        cv = @counts[n]
      end
      totalv += cv
      case which
      when PRboth
        of.printf(pc ? "%.6f %.2f %.2f\n" : "%.6f %d %d\n",
          x*@xscale, cv, totalv) if cv != 0
      when PRxonly
        of.printf(" %.#{pc}f", x*@xscale)
      when PRxreals
        of.push(x*@xscale)
      when PRyonly
        of.printf(pc ? " %.3f" : " %d", cv)
      when PRcumuly
        of.printf(pc ? " %.3f" : " %d", totalv)
      end
    end
    of  # For PRxreals
  end

  def dump(of, pc)
    write(of, pc, PRboth)
  end

  def write_x(of, nfd)  # Nbr of %f digits
    write(of, nfd, PRxonly)
    of.print "\n"
  end

  def write_xreals
    of = Array.new
    xx = write(of, false, PRxreals)
#    print "xx=#{xx}\n"
  end

  def write_y(of, pc)
    write(of, pc, PRyonly)
    of.print "\n"
  end

  def write_cumul_y(of, pc)
    write(of, pc, PRcumuly)
    of.print "\n"
  end

  def dump_csv(of, pc)  # Just dump distribution (with its parameters)
    of.printf("%.5f, %.5f, %d", @lower, @upper, @bins)
    @counts.each do |c|
      of.print ", #{v = pc ? sprintf("%0.3f", c*100.0/@n_counts) : c}"
    end
  end
end  # Distribution
