#!/usr/bin/env ruby

#  2015, Tue  1 Jun 10 (NZST)
#  1622, Tue 13 Apr 10 (PDT)
#  1425, Thu 31 Dec 09 (NZDT)
#
#  owtmon.rb: OWT monitor, with Ruby source classifier
#
#  Copyright 2007-2010, Nevil Brownlee, U Auckland | CAIDA | WAND
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "#{File.dirname($0)}/owt-config"

require "#{Home}/IPprefix" 
require "#{Home}/owt-c/OWTmonitor"  # Must follow IPprefix (!)
require "#{Home}/owt-rb/owt-support" 
require "#{Home}/owt-rb/owt-monitor" 
require "#{Home}/owt-rb/write-sum" 
require "#{Home}/owt-rb/distrib"

# ./owtmon.rb dag:/dev/dag0 sum2                                  # ua-live
# ./owtmon.rb traces/ua.20091201.1700.bpf.gz sum2 sum2/san-trace  # ua-trace
# ./owtmon.rb /work/telescope/live/2009-12-21/net44.1261404000.pcap.gz sum2 sum2/test/2009-12-21.fdf  # odin

trace_URI = ARGV[0]    # URI of trace to process
output_dn = ARGV[1]    # output sum2 directory 
log_fn = ARGV[2]       # name of owtmon log file (appended to)

$clog_level = ARGV.size < 4 ? 3 : ARGV[3].to_i
  # 0 or empty = display all log messages
  # 1          = display Time messages
  # 2          = display Time and Stats messages


trace_URI = SourceURI  if trace_URI.nil?
output_dn = SummaryDir if output_dn.nil?
log_fn = StatsFile     if log_fn.nil?

OWTmonVer = '1.6';  FormatVer = '1.0'

TCP = 6;  UDP = 17;  ICMP = 1;  ICMP6 = 58  # Protocol numbers
ICMP_DEST_UNREACH  =   3  # ICMP Types
ICMP_TTL_EXCEEDED  =  11
ICMP6_DEST_UNREACH =   1  # ICMP6 Types
ICMP6_TTL_EXCEEDED =   3
# TCP flags  32 URG, 16 ACK, 8 PSH, 4 RST, 2 SYN, 1 FIN 

def file_dn(output_dn, ts)
  check_dir_exists(output_dn)
  odn = "#{output_dn}/#{utc_hts(ts)}"
  check_dir_exists(odn)
  odn
end

def day_change(ts)
  if $today_date.nil?
    $today_date = utc_hts(ts)
    return true
  else
    new_date = utc_hts(ts)
    change = new_date != $today_date
    $today_date = new_date
    return change
  end
end

def next_uri(uri)  # Get filename for next telescope trace file
  uri =~ /^(.+):(.+)$/
  utype = $1;  fn = $2
  arg_fn = File.basename(fn)
  
  fn =~ /^(.+)\/(20.+)\/(.+)\.(\d+)\.(.+)$/
  live_dir = $1;  sdate = $2;  net = $3;  ut = $4;  ftype = $5

  return_next = false
  Dir.glob("#{live_dir}/????-??-??").sort.each do |an|
    date = File.basename(an) 
    next if date < sdate
    Dir.glob("#{live_dir}/#{date}/#{net}.*.#{ftype}").sort.each do |fulln| 
      fn = File.basename(fulln)
      return "#{utype}:#{fulln}" if return_next
      return_next = true if fn == arg_fn
    end
  end
end

class NbrHash < Hash  # For port numbers and ip lengths
  def update(str)
    if !self[str]
      self[str] = 1
      return true  # new
    else
      self[str] += 1
      return false
    end
  end

  def val_range
    min = 65535;  max = 1
    self.each do |key, val|
      ikey = key.to_i
      min = ikey if ikey < min
      max = ikey if ikey > max
    end
    [self.size, min, max]
  end

  def mode
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    shash[0][0].to_i
  end

  def write(f)
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    n = 0
    shash.each do |kv|
      f.printf("  %u %u", kv[0], kv[1])
      n += 1
      break if n == N_ProbePorts
    end
  end
end  # NbrHash

class RetryHash < Hash  # For pkt_keys, to get average retry timeout
  def initialize
    super  # Initialise the Hash
    @n_retries = 0;  @tot_retry_time = 0.0
  end

  def update(key, arrival_time)
    last_a_t = self[key]
    if !last_a_t  # New key
      self[key] = arrival_time
      return true  # new
    else
      @tot_retry_time += arrival_time - last_a_t
      @n_retries += 1
      self[key] = arrival_time
      return false
    end
  end

  def mean_retry_time
    if @n_retries == 0
      nil
    else
      @tot_retry_time/@n_retries
    end
  end
end  # RetryHash

$pp_count = 0  # Count of sources with all pkts to single port

class SourceHost
  def initialize(owt_event)  # _ENOUGH or _END_OF_FILE
    @event = owt_event
    @first_fwd, @src_addr, @n_recs, @lifetime = owt_event.usage
#d printf("initialise: %s  %d  %.3f\n", @src_addr, @n_recs, @lifetime)
    @pkts = RetryHash.new
    @dst_addrs = NbrHash.new  
    @src_ports = NbrHash.new;  @dst_ports = NbrHash.new
    @pkt_lens = NbrHash.new
    @n_pkts = @n_tcp = @n_udp = @n_bscat = @n_icmp = @n_other = 0
    @cfc_count = 0
  end

  def update(pkt)  # pkt is an array  of packet attributes
    pkt_key, arrival_time, proto, src_port, dst_port, dst_addr,
      ip_len, tcp_flags, cfc = pkt
#printf("%2d  %.6f  %u  %u  %u  %s  %d  %d  %d:  ",
#    @n_pkts, arrival_time, proto, src_port, dst_port, dst_addr,
#      ip_len, tcp_flags, cfc)
    @n_pkts += 1

    new = @pkts.update(pkt_key, arrival_time)
    @pkt_lens.update(ip_len)
    @cfc_count += 1 if new && cfc == 1
    @dst_addrs.update(dst_addr.to_s)
    case proto
      when TCP
        @n_tcp += 1
        @src_ports.update(src_port)
        @dst_ports.update(dst_port)
        @n_bscat += 1 if tcp_flags == 18 || tcp_flags == 20
          # TCP ACK+SYNs or ACK+RSTs
      when UDP
        @n_udp += 1
        @src_ports.update(src_port)
        @dst_ports.update(dst_port)
      # For ICMP and ICMP6 pkts, tcp_flags has the ICMP Type
      when ICMP
        if tcp_flags == ICMP_TTL_EXCEEDED || tcp_flags == ICMP_DEST_UNREACH
          @n_bscat += 1
        else
          @n_icmp += 1
        end
      when ICMP6
        if tcp_flags == ICMP6_TTL_EXCEEDED || tcp_flags == ICMP6_DEST_UNREACH
          @n_bscat += 1
        else
          @n_icmp += 1
        end
      else
        @n_other += 1
    end
#   printf("%s:  %s  %d  %d %d %d %d\n", 
#     $analyse_event, new, @cfc_count, @n_tcp, @n_udp, @n_icmp, @n_other)
  end

  def analyse  # (tsum)
#ua    print "#{$analyse_event}: pkt_lens.size=>#{@pkt_lens.size}<\n"
    return if @n_pkts == 0
      # No packets seen since source timed out, no new info

    n_ports, min_port, max_port = @dst_ports.val_range
    probe_port = nil

    if @n_pkts < MinPkts
      # Format 1.1, 2010-02-05: classify sources that only
      #   send TCP or UDP probe packets
      # Reconstructed 2010-02-16
      st = T_unclassified
      if @n_tcp == @n_pkts 
        if n_ports == 1 && @dst_addrs.size > @pkts.size/2
          st = T_tcp_probe;  probe_port = min_port
        end
      elsif @n_udp == @n_pkts
        if n_ports == 1 && @dst_addrs.size > @pkts.size/2
          st = T_udp_probe;  probe_port = min_port
        end
      end
      @event.set_source_attribs( [st, probe_port,
        @pkts.mean_retry_time, @dst_ports.size, @pkt_lens.mode] )
      return
    end

    if  @cfc_count != 0 && @n_icmp+@n_other == 0
      if @cfc_count == @pkts.size
        st = T_conficker_C
      else
        st = T_conficker_mix  # Some pkts not cf_p2p
      end
    else
      if @n_bscat == @n_pkts
        st = T_backscatter
      elsif @n_tcp == @n_pkts
        if n_ports == 1 && @dst_addrs.size > @pkts.size/2
          st = T_tcp_probe;  probe_port = min_port
        elsif @dst_addrs.size == @pkts.size 
          st = T_tcp_single_ppa
        else
          st = T_tcp_multi_ppa
        end
      elsif @n_udp == @n_pkts
        if n_ports == 1 && @dst_addrs.size > @pkts.size/2
          st = T_udp_probe;  probe_port = min_port
        elsif @dst_addrs.size == @pkts.size 
          st = T_udp_single_ppa
        else
          st = T_udp_multi_ppa
        end
      elsif @n_icmp+@n_other == 0
        st = T_mixed_tcp_udp
      else
        st = T_no_tcp_or_udp;  n_ports = 0
      end
    end

  @event.set_source_attribs( [st, probe_port,
    @pkts.mean_retry_time, @dst_ports.size, @pkt_lens.mode] )
  end
end  # SourceHost

class OWTevent
  attr_reader :id
end

class OWTmessage
  attr_reader :type, :source, :text
end

# sudo dagfour -d /dev/dag0 -aei
# sudo dagfour slen=80 varlen
# sudo dagfour -d /dev/dag0 disableb
# sudo dagfour -d /dev/dag0 enablea  # A interface sees UA border traffic
# sudo dagsnap -v -o tracefile  # Check it's working

Type_s = {OQ_TXTMSG => 'OQ_TXTMSG', 
  OQ_STARTED => 'OQ_STARTED', OQ_PROGRESS => 'OQ_PROGRESS',
  OQ_ENOUGH => 'OQ_ENOUGH', OQ_TIMED_OUT => 'OQ_TIMED_OUT',
  OQ_END_OF_FILE => 'OQ_END_OF_FILE' }

$stderr.print \
  "trace_URI=#{trace_URI}, output_dn=#{output_dn}, log_fn=#{log_fn}\n"

$owt = Monitor.new("-i#{trace_URI} -s#{N_Streams} -p#{N_PktDatas} -r#{EnoughPkts} -n#{MinPkts} -t#{N_source_types} -c#{Hours_reqd*3600} -h#{Hours_reqd}")
addr_list = Array.new
LocalAddress.split(/,\s*/).each { |la| addr_list.push(IPprefix.from_s(la)) }
$owt.set_local_addrs( addr_list )

(Meter_ver, TwoWay) = $owt.info
print "Meter_ver=#{Meter_ver}, TwoWay=#{TwoWay}\n"

LiveTrace = $owt.live
if LiveTrace
  Trace_sz = Host  # Dummy for trace from live interface
else
  if $owt.n_interfaces != 1
    $stderr.print "Can only process one file per owtmon.rb run!\n"
    exit
  else
    Trace_sz = sprintf("%.3f", File.size($owt.last_fn)/1000000000.0)  # GB
  end
end

def sigint
  $stderr.print "^C pressed: stopping owtmon ...\n"
  $owt.close
end

trap(:INT, lambda {sigint})  # ^C (Interrupt) =>  Set request_shutdown

#def sigtstp
#  $stderr.print "SIGTSTP (^Z pressed) ..\n"
#  st = caller(3)  # Ignore first 3 lines
#    # First line printed is where the signal found us
#  $stderr.print st.join("\n")
#  $stderr.print "\n"
#end 
##trap(:TSTP, lambda {sigtstp})  # ^Z (Temporary stop) => show traceback

trap(:TSTP, lambda {} )  # Ignore ^Z (used to stop mk-hourly-sum2.rb)
  # NOTE: ^Z won't push owtmon.rb to background - ask it to stop with ^C
  #       or force it to stop with KILL -9 (after ps to get its procid)

$start_proc_time = Process::times;  $start_proc_date = Time.now

np_dist = nil

MkStrmSzDistribs = false  # Don't write stream-size distribs
szf = File.new('size.dat', 'w') if MkStrmSzDistribs

if Hours_reqd != 1
  uri = "pcapfile:#{trace_fn}"
  nu = 0;  other_uris = Array.new
  (1..Hours_reqd-1).each do
    new_uri = next_uri(uri)
    other_uris.push(new_uri)
    uri = new_uri
  end
  $owt.set_uris(other_uris)
end

n = 0
first_time = true
sum_start_ts = $owt.open
if LiveTrace
  log_fn = "#{file_dn(output_dn, sum_start_ts)}/#{utc_gts(sum_start_ts)}.fdf"
end
$fdf = File.open(log_fn, 'a')

$owt.each do |msg|
  # print "#{Type_s[msg.type]} ---\n"
  case msg.type
  when OQ_TXTMSG
    type = msg.text[1..4]
    $fdf.print "#{msg.text}\n"  # Always written to log file
    exit if msg.text =~ /Fatal/
    mts = msg.source  # g_last_ts for TXTMSGs
    case type  # txt message type
    when 'Time'  #Time
      if $clog_level > 0
        print "#{msg.text}\n";  $stdout.flush
      end
      if MkStrmSzDistribs
        szf.print "#{utc_gts(mts)} "
        if first_time
          np_dist = Distribution.new(1, 400, 100, DT_LOG, 1.0)
          np_dist.write_x(szf, false)
          first_time = false
        else
          $owt.each_stream_size do |sz|
            np_dist.count(sz)
          end
          np_dist.write_y(szf, false)
          np_dist = Distribution.new(1, 400, 100, DT_LOG, 1.0)
        end
      end
    when 'Reco'  # Record: time to write sum2 file
      fn = "#{file_dn(output_dn, sum_start_ts)}" +
        "/#{utc_gts(sum_start_ts)}.#{SumFileType}"
      write_sum_file(fn, sum_start_ts, mts) 
      if day_change(mts)  # Start new log file
        log_fn = "#{file_dn(output_dn, mts)}/#{utc_gts(mts)}.fdf"
        $fdf.close
        $fdf = File.open(log_fn, 'a')
      end
      sum_start_ts = mts
    when 'Stat'
      if $clog_level > 1
        print "#{msg.text}\n";  $stdout.flush
      end
    when 'Shut'  #Shutdown
      "#{msg.text}\n"  # Always written to console
      if MkStrmSzDistribs
        szf.print "#{utc_gts(mts)} "
        $owt.each_stream_size do |sz|
          np_dist.count(sz)
        end
        np_dist.write_y(szf, false)
      end

      # By now, all the EOF streams have been classified
      fn = "#{file_dn(output_dn, sum_start_ts)}" +
        "/#{utc_gts(sum_start_ts)}.#{SumFileType}"
      write_sum_file(fn, sum_start_ts, mts) 
      break  # End of .each loop
    else  # Shouldn't happen
      print "#{msg.text}\n"  # Always written to console
    end

  when OQ_ENOUGH, OQ_END_OF_FILE,
      OQ_TIMED_OUT  # One-way flow has timed out
    event = msg.source
    s = SourceHost.new(event)
    event.each_pkt do |p|
      s.update(p)
    end
$analyse_event = msg.type
    s.analyse

    # $fdf.print "#{Type_s[msg.type]}:  "
    # first_fwd, src_addr, tot_pkts, lifetime = event.usage
    # inward = first_fwd ? '->' : '<-'
    # $fdf.printf("%s %s  %d  %.6f\n",
    #   src_addr, inward, tot_pkts, lifetime)

  else  # Shouldn't happen !
    event = msg.source
    $fdf.print "#{Type_s[msg.type]}:  "
    first_fwd, src_addr, tot_pkts, lifetime = event.usage
    inward = first_fwd ? '->' : '<-'
    $fdf.printf("*** %s %s  %d  %.6f\n", src_addr, inward, tot_pkts, lifetime)
    exit
  end
#  n += 1
#    exit if n == 5
end

#szf.close if MkStrmSzDistribs

$fdf.close
